"""
It's common practice in Linux distributions to debundle vendored
Python packages and modules for many reasons. Usually it's done
by direct patching of imports of vendored packages. It's can be
very time-consuming to rediff such patches on new upstream release.

This module transparently proxies imports of bundled top level
module/package to a normal one (any imported package with that name)
according to PEP302 rules. Thereby, it eliminates the need for source
code patching.

How to use:
- don't modify any source files to fix imports
- remove top level directory for vendored packages
- put this file as __init__.py into that newly created root-directory
  and configure replaceable variables


For example, six and requests are vendored into path mypackage/vendored.
Then changes for RPM spec file would look like:
- filter no longer provided self-dependencies
  %filter_from_requires /python3(mypackage\.vendored\..*)/d

- add new dependencies on debundled packages
  %py3_requires six
  %py3_requires requests

- bundle this module :)
  VENDORED_PATH='mypackage/vendored'
  UNVENDORED_PATH="$VENDORED_PATH/__init__.py"
  rm -r "$VENDORED_PATH"
  mkdir "$VENDORED_PATH"
  cp "%SOURCE1" "$UNVENDORED_PATH"
  sed -i \
        -e 's/"botocore.vendored"/"mypackage.vendored"/' \
        -e 's/{"requests.packages"}/None/' \
        "$UNVENDORED_PATH"

Hooray, we are done.

Checkout for a newer version:
https://git.altlinux.org/people/slev/public/debundler.git
"""
import sys
import importlib
import importlib.util


class UnvendoredImporter:
    def __init__(self, root, fake_packages=None):
        """
        root defines a root of bundled packages. For example,
        if some project bundles other packages into 'package/vendored/'
        then the root would be 'package.vendored'.

        fake_packages defines modules that behave like package,
        for example, requests.packages dirty hacks an import system
        (import packages into module and add them into its namespace).
        """
        self.roots = {root if root.endswith(".") else root + "."}
        self.fake_packages = set()
        if fake_packages is not None:
            self.fake_packages.update(set(fake_packages))

    def install(self):
        for importer in sys.meta_path:
            if importer.__class__.__name__ == self.__class__.__name__:
                importer.add(self)
                break
        else:
            sys.meta_path.insert(0, self)

    def add(self, other):
        self.roots.update(other.roots)
        self.fake_packages.update(other.fake_packages)

    def _parse_name(self, fullname):
        for root in self.roots:
            vendored_root, _, vendored_pkg = fullname.partition(root)
            # start with root
            if not vendored_root:
                return vendored_root, vendored_pkg

        return None, None

    def create_module(self, spec):
        fullname = spec.name
        vendored_root, vendored_pkg = self._parse_name(fullname)
        if vendored_root is not None:
            mod = importlib.import_module(vendored_pkg)
            sys.modules[fullname] = mod
            return mod

        return None

    def exec_module(self, module):
        pass

    def is_package(self, fullname):
        if not self.fake_packages:
            return None

        vendored_root, vendored_pkg = self._parse_name(fullname)
        if vendored_root is not None:
            # module that behave like package
            if vendored_pkg in self.fake_packages:
                return True

        return None

    def find_spec(self, fullname, path, target=None):
        if path is None:
            return None

        if fullname.startswith(tuple(self.roots)):
            return importlib.util.spec_from_loader(fullname, self)

        return None


UnvendoredImporter(
    "botocore.vendored",
    fake_packages={"requests.packages"},
).install()
